/* Copyright Airship and Contributors */

#import "UAPreferenceDataStore+Internal.h"
#import "UAGlobal.h"

@interface UAPreferenceDataStore()
@property (nonatomic, strong) NSUserDefaults *defaults;
@property (nonatomic, copy) NSString *keyPrefix;
@end


@implementation UAPreferenceDataStore

+ (instancetype)preferenceDataStoreWithKeyPrefix:(NSString *)keyPrefix {
    UAPreferenceDataStore *dataStore = [[UAPreferenceDataStore alloc] init];
    dataStore.defaults = [self createDefaultsWithPrefix:keyPrefix];
    dataStore.keyPrefix = keyPrefix;
    return dataStore;
}

+ (NSUserDefaults *)createDefaultsWithPrefix:(NSString *)keyPrefix {
    NSString *bundleID = NSBundle.mainBundle.bundleIdentifier ?: @"";
    NSString *suiteName = [NSString stringWithFormat:@"%@.airship.settings", bundleID];
    
    NSUserDefaults *defaults = [[NSUserDefaults alloc] initWithSuiteName:suiteName];
    if (!defaults) {
        UA_LERR(@"Failed to create defaults %@", suiteName);
        return NSUserDefaults.standardUserDefaults;
    }
    
    NSDictionary *defaultDictionary = [NSUserDefaults.standardUserDefaults dictionaryRepresentation];
    
    for (NSString *key in defaultDictionary.allKeys) {
        if ([key hasPrefix:keyPrefix]) {
            [defaults setValue:defaultDictionary[key] forKey:key];
            [NSUserDefaults.standardUserDefaults removeObjectForKey:key];
        }
    }
    
    return defaults;
}

- (NSString *)prefixKey:(nonnull NSString *)key {
    return [self.keyPrefix stringByAppendingString:key];
}

- (id)valueForKey:(NSString *)key {
    return [self.defaults valueForKey:[self prefixKey:key]];
}

- (void)setValue:(id)value forKey:(NSString *)key {
    [self.defaults setValue:value forKey:[self prefixKey:key]];
}

- (void)removeObjectForKey:(NSString *)key {
    [self.defaults removeObjectForKey:[self prefixKey:key]];
}

- (BOOL)keyExists:(NSString *)key {
    return ([self objectForKey:key] != nil);
}

- (id)objectForKey:(NSString *)key {
    return [self.defaults objectForKey:[self prefixKey:key]];
}

- (NSString *)stringForKey:(NSString *)key {
    return [self.defaults stringForKey:[self prefixKey:key]];
}

- (NSArray *)arrayForKey:(NSString *)key {
    return [self.defaults arrayForKey:[self prefixKey:key]];
}

- (NSDictionary *)dictionaryForKey:(NSString *)key {
    return [self.defaults dictionaryForKey:[self prefixKey:key]];
}

- (NSData *)dataForKey:(NSString *)key {
    return [self.defaults dataForKey:[self prefixKey:key]];
}

- (NSArray *)stringArrayForKey:(NSString *)key {
    return [self.defaults stringArrayForKey:[self prefixKey:key]];
}

- (NSInteger)integerForKey:(NSString *)key {
    return [self.defaults integerForKey:[self prefixKey:key]];
}

- (float)floatForKey:(NSString *)key {
    return [self.defaults floatForKey:[self prefixKey:key]];
}

- (double)doubleForKey:(NSString *)key {
    return [self.defaults doubleForKey:[self prefixKey:key]];
}

- (double)doubleForKey:(NSString *)key defaultValue:(double)defaultValue {
    if ([self keyExists:key]) {
        return [self doubleForKey:key];
    } else {
        return defaultValue;
    }
}

- (BOOL)boolForKey:(NSString *)key {
    return [self.defaults boolForKey:[self prefixKey:key]];
}

- (BOOL)boolForKey:(NSString *)key defaultValue:(BOOL)defaultValue {
    if ([self keyExists:key]) {
        return [self boolForKey:key];
    } else {
        return defaultValue;
    }
}

- (NSURL *)URLForKey:(NSString *)key {
    return [self.defaults URLForKey:[self prefixKey:key]];
}

- (void)setInteger:(NSInteger)value forKey:(NSString *)key {
    [self.defaults setInteger:value forKey:[self prefixKey:key]];
}

- (void)setFloat:(float)value forKey:(NSString *)key {
    [self.defaults setFloat:value forKey:[self prefixKey:key]];
}

- (void)setDouble:(double)value forKey:(NSString *)key {
    [self.defaults setDouble:value forKey:[self prefixKey:key]];
}

- (void)setBool:(BOOL)value forKey:(NSString *)key {
    [self.defaults setBool:value forKey:[self prefixKey:key]];
}

- (void)setURL:(NSURL *)value forKey:(NSString *)key {
    [self.defaults setURL:value forKey:[self prefixKey:key]];
}

- (void)setObject:(id)value forKey:(NSString *)key {
    [self.defaults setObject:value forKey:[self prefixKey:key]];
}

- (void)migrateUnprefixedKeys:(NSArray *)keys {
    // No-Op
}

- (void)removeAll {
    for (NSString *key in [[self.defaults dictionaryRepresentation] allKeys]) {
        if ([key hasPrefix:self.keyPrefix]) {
            [self.defaults removeObjectForKey:key];
        }
    }
}

@end
